@@ -157,9 +157,6 @@ public class AquaKeyBindings {
                         "shift alt KP_LEFT", null,
                         "shift alt RIGHT", null,
                         "shift alt KP_RIGHT", null,
-                        "alt BACK_SPACE", null,
-                        "ctrl W", null,
-                        "alt DELETE", null,
                 }));
     }
 
@@ -389,7 +386,11 @@ public class AquaKeyBindings {
             "shift ENTER", "selectPreviousRowCell",
             "alt TAB", "focusHeader",
             "alt shift TAB", "focusHeader",
-            "F8", "focusHeader"
+            "F8", "focusHeader",
+            "ctrl shift UP", "selectFirstRowExtendSelection",
+            "ctrl shift DOWN", "selectLastRowExtendSelection",
+            "ctrl shift RIGHT", "selectLastColumnExtendSelection",
+            "ctrl shift LEFT", "selectFirstColumnExtendSelection"
         }));
     }
 
