@@ -3,9 +3,9 @@ package com.sun.tools.jdeps.resources;
 public final class version extends java.util.ListResourceBundle {
     protected final Object[][] getContents() {
         return new Object[][] {
-            { "full", "21.0.10+8-LTS-jvmci-23.1-b84" },
-            { "jdk", "21.0.10" },
-            { "release", "21.0.10" },
+            { "full", "24.0.2+11-jvmci-b01" },
+            { "jdk", "24.0.2" },
+            { "release", "24.0.2" },
         };
     }
 }
