@@ -24,78 +24,70 @@
  */
 package com.oracle.svm.core.graal.snippets;
 
-import static com.oracle.svm.core.SubstrateOptions.MultiThreaded;
-import static com.oracle.svm.core.SubstrateOptions.SpawnIsolates;
 import static com.oracle.svm.core.graal.nodes.WriteCurrentVMThreadNode.writeCurrentVMThread;
 import static com.oracle.svm.core.graal.nodes.WriteHeapBaseNode.writeCurrentVMHeapBase;
 import static com.oracle.svm.core.heap.RestrictHeapAccess.Access.NO_ALLOCATION;
+import static com.oracle.svm.core.imagelayer.ImageLayerSection.SectionEntries.HEAP_BEGIN;
 import static com.oracle.svm.core.util.VMError.shouldNotReachHereUnexpectedInput;
+import static jdk.graal.compiler.core.common.spi.ForeignCallDescriptor.CallSideEffect.HAS_SIDE_EFFECT;
 
 import java.util.Map;
 
-import org.graalvm.compiler.api.replacements.Fold;
-import org.graalvm.compiler.api.replacements.Snippet;
-import org.graalvm.compiler.core.common.CompressEncoding;
-import org.graalvm.compiler.core.common.spi.ForeignCallDescriptor;
-import org.graalvm.compiler.graph.Node;
-import org.graalvm.compiler.graph.Node.ConstantNodeParameter;
-import org.graalvm.compiler.graph.Node.NodeIntrinsic;
-import org.graalvm.compiler.nodes.PauseNode;
-import org.graalvm.compiler.nodes.extended.ForeignCallNode;
-import org.graalvm.compiler.nodes.spi.LoweringTool;
-import org.graalvm.compiler.options.OptionValues;
-import org.graalvm.compiler.phases.util.Providers;
-import org.graalvm.compiler.replacements.SnippetTemplate;
-import org.graalvm.compiler.replacements.SnippetTemplate.Arguments;
-import org.graalvm.compiler.replacements.SnippetTemplate.SnippetInfo;
-import org.graalvm.compiler.replacements.Snippets;
-import org.graalvm.compiler.word.Word;
 import org.graalvm.nativeimage.CurrentIsolate;
 import org.graalvm.nativeimage.ImageSingletons;
 import org.graalvm.nativeimage.Isolate;
 import org.graalvm.nativeimage.IsolateThread;
 import org.graalvm.nativeimage.StackValue;
+import org.graalvm.nativeimage.c.struct.SizeOf;
 import org.graalvm.nativeimage.c.type.CCharPointer;
 import org.graalvm.nativeimage.c.type.CLongPointer;
 import org.graalvm.nativeimage.c.type.WordPointer;
 import org.graalvm.word.LocationIdentity;
+import org.graalvm.word.Pointer;
 import org.graalvm.word.PointerBase;
 import org.graalvm.word.UnsignedWord;
 import org.graalvm.word.WordFactory;
 
 import com.oracle.svm.core.CPUFeatureAccess;
 import com.oracle.svm.core.IsolateArgumentParser;
+import com.oracle.svm.core.IsolateArguments;
 import com.oracle.svm.core.IsolateListenerSupport;
 import com.oracle.svm.core.Isolates;
 import com.oracle.svm.core.JavaMainWrapper.JavaMainSupport;
 import com.oracle.svm.core.NeverInline;
 import com.oracle.svm.core.RuntimeAssertionsSupport;
 import com.oracle.svm.core.SubstrateDiagnostics;
-import com.oracle.svm.core.SubstrateGCOptions;
 import com.oracle.svm.core.SubstrateOptions;
 import com.oracle.svm.core.SubstrateUtil;
 import com.oracle.svm.core.Uninterruptible;
+import com.oracle.svm.core.UnmanagedMemoryUtil;
 import com.oracle.svm.core.c.CGlobalData;
 import com.oracle.svm.core.c.CGlobalDataFactory;
+import com.oracle.svm.core.c.NonmovableArrays;
 import com.oracle.svm.core.c.function.CEntryPointActions;
 import com.oracle.svm.core.c.function.CEntryPointCreateIsolateParameters;
 import com.oracle.svm.core.c.function.CEntryPointErrors;
 import com.oracle.svm.core.c.function.CEntryPointNativeFunctions;
-import com.oracle.svm.core.c.function.CEntryPointSetup;
+import com.oracle.svm.core.c.locale.LocaleSupport;
 import com.oracle.svm.core.code.CodeInfoTable;
+import com.oracle.svm.core.config.ConfigurationValues;
+import com.oracle.svm.core.container.Container;
 import com.oracle.svm.core.graal.meta.SubstrateForeignCallsProvider;
 import com.oracle.svm.core.graal.nodes.CEntryPointEnterNode;
 import com.oracle.svm.core.graal.nodes.CEntryPointLeaveNode;
 import com.oracle.svm.core.graal.nodes.CEntryPointUtilityNode;
-import com.oracle.svm.core.graal.nodes.WriteCurrentVMThreadNode;
+import com.oracle.svm.core.heap.Heap;
 import com.oracle.svm.core.heap.PhysicalMemory;
 import com.oracle.svm.core.heap.ReferenceHandler;
 import com.oracle.svm.core.heap.ReferenceHandlerThread;
 import com.oracle.svm.core.heap.RestrictHeapAccess;
+import com.oracle.svm.core.imagelayer.ImageLayerBuildingSupport;
+import com.oracle.svm.core.imagelayer.ImageLayerSection;
 import com.oracle.svm.core.jdk.PlatformNativeLibrarySupport;
 import com.oracle.svm.core.jdk.RuntimeSupport;
 import com.oracle.svm.core.log.Log;
 import com.oracle.svm.core.option.RuntimeOptionParser;
+import com.oracle.svm.core.os.CommittedMemoryProvider;
 import com.oracle.svm.core.os.MemoryProtectionProvider;
 import com.oracle.svm.core.os.VirtualMemoryProvider;
 import com.oracle.svm.core.snippets.SnippetRuntime;
@@ -105,13 +97,32 @@ import com.oracle.svm.core.stack.StackOverflowCheck;
 import com.oracle.svm.core.thread.PlatformThreads;
 import com.oracle.svm.core.thread.Safepoint;
 import com.oracle.svm.core.thread.ThreadListenerSupport;
+import com.oracle.svm.core.thread.ThreadingSupportImpl;
 import com.oracle.svm.core.thread.VMOperationControl;
 import com.oracle.svm.core.thread.VMThreads;
 import com.oracle.svm.core.thread.VMThreads.SafepointBehavior;
-import com.oracle.svm.core.threadlocal.VMThreadLocalMTSupport;
+import com.oracle.svm.core.threadlocal.VMThreadLocalSupport;
 import com.oracle.svm.core.util.UnsignedUtils;
 import com.oracle.svm.core.util.VMError;
 
+import jdk.graal.compiler.api.replacements.Fold;
+import jdk.graal.compiler.api.replacements.Snippet;
+import jdk.graal.compiler.core.common.CompressEncoding;
+import jdk.graal.compiler.core.common.spi.ForeignCallDescriptor;
+import jdk.graal.compiler.graph.Node;
+import jdk.graal.compiler.graph.Node.ConstantNodeParameter;
+import jdk.graal.compiler.graph.Node.NodeIntrinsic;
+import jdk.graal.compiler.nodes.NamedLocationIdentity;
+import jdk.graal.compiler.nodes.PauseNode;
+import jdk.graal.compiler.nodes.extended.ForeignCallNode;
+import jdk.graal.compiler.nodes.spi.LoweringTool;
+import jdk.graal.compiler.options.OptionValues;
+import jdk.graal.compiler.phases.util.Providers;
+import jdk.graal.compiler.replacements.SnippetTemplate;
+import jdk.graal.compiler.replacements.SnippetTemplate.Arguments;
+import jdk.graal.compiler.replacements.SnippetTemplate.SnippetInfo;
+import jdk.graal.compiler.replacements.Snippets;
+import jdk.graal.compiler.word.Word;
 import jdk.internal.misc.Unsafe;
 
 /**
@@ -124,23 +135,32 @@ import jdk.internal.misc.Unsafe;
  */
 public final class CEntryPointSnippets extends SubstrateTemplates implements Snippets {
 
-    public static final SubstrateForeignCallDescriptor CREATE_ISOLATE = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "createIsolate", false, LocationIdentity.any());
-    public static final SubstrateForeignCallDescriptor INITIALIZE_ISOLATE = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "initializeIsolate", false, LocationIdentity.any());
-    public static final SubstrateForeignCallDescriptor ATTACH_THREAD = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "attachThread", false, LocationIdentity.any());
-    public static final SubstrateForeignCallDescriptor ENSURE_JAVA_THREAD = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "ensureJavaThread", false, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor CREATE_ISOLATE = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "createIsolate",
+                    HAS_SIDE_EFFECT, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor INITIALIZE_ISOLATE = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "initializeIsolate",
+                    HAS_SIDE_EFFECT, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor ATTACH_THREAD = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "attachThread", HAS_SIDE_EFFECT,
+                    LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor ENSURE_JAVA_THREAD = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "ensureJavaThread",
+                    HAS_SIDE_EFFECT, LocationIdentity.any());
 
-    public static final SubstrateForeignCallDescriptor ENTER_BY_ISOLATE_MT = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "enterByIsolateMT", false, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor ENTER_BY_ISOLATE = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "enterByIsolate",
+                    HAS_SIDE_EFFECT, LocationIdentity.any());
 
-    public static final SubstrateForeignCallDescriptor DETACH_THREAD_MT = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "detachThreadMT", false, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor DETACH_CURRENT_THREAD = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "detachCurrentThread",
+                    HAS_SIDE_EFFECT, LocationIdentity.any());
 
-    public static final SubstrateForeignCallDescriptor REPORT_EXCEPTION = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "reportException", false, LocationIdentity.any());
-    public static final SubstrateForeignCallDescriptor TEAR_DOWN_ISOLATE = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "tearDownIsolate", false, LocationIdentity.any());
-    public static final SubstrateForeignCallDescriptor IS_ATTACHED_MT = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "isAttachedMT", false, LocationIdentity.any());
-    public static final SubstrateForeignCallDescriptor FAIL_FATALLY = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "failFatally", false, LocationIdentity.any());
-    public static final SubstrateForeignCallDescriptor VERIFY_ISOLATE_THREAD = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "verifyIsolateThread", false, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor REPORT_EXCEPTION = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "reportException",
+                    HAS_SIDE_EFFECT, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor TEAR_DOWN_ISOLATE = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "tearDownIsolate",
+                    HAS_SIDE_EFFECT, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor IS_ATTACHED = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "isAttached", HAS_SIDE_EFFECT, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor FAIL_FATALLY = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "failFatally", HAS_SIDE_EFFECT, LocationIdentity.any());
+    public static final SubstrateForeignCallDescriptor VERIFY_ISOLATE_THREAD = SnippetRuntime.findForeignCall(CEntryPointSnippets.class, "verifyIsolateThread", HAS_SIDE_EFFECT,
+                    LocationIdentity.any());
 
-    public static final SubstrateForeignCallDescriptor[] FOREIGN_CALLS = {CREATE_ISOLATE, INITIALIZE_ISOLATE, ATTACH_THREAD, ENSURE_JAVA_THREAD, ENTER_BY_ISOLATE_MT,
-                    DETACH_THREAD_MT, REPORT_EXCEPTION, TEAR_DOWN_ISOLATE, IS_ATTACHED_MT, FAIL_FATALLY, VERIFY_ISOLATE_THREAD};
+    public static final SubstrateForeignCallDescriptor[] FOREIGN_CALLS = {CREATE_ISOLATE, INITIALIZE_ISOLATE, ATTACH_THREAD, ENSURE_JAVA_THREAD, ENTER_BY_ISOLATE,
+                    DETACH_CURRENT_THREAD, REPORT_EXCEPTION, TEAR_DOWN_ISOLATE, IS_ATTACHED, FAIL_FATALLY, VERIFY_ISOLATE_THREAD};
 
     @NodeIntrinsic(value = ForeignCallNode.class)
     public static native int runtimeCall(@ConstantNodeParameter ForeignCallDescriptor descriptor, CEntryPointCreateIsolateParameters parameters);
@@ -151,6 +171,9 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
     @NodeIntrinsic(value = ForeignCallNode.class)
     public static native int runtimeCall(@ConstantNodeParameter ForeignCallDescriptor descriptor, Isolate isolate);
 
+    @NodeIntrinsic(value = ForeignCallNode.class)
+    public static native int runtimeCall(@ConstantNodeParameter ForeignCallDescriptor descriptor);
+
     @NodeIntrinsic(value = ForeignCallNode.class)
     public static native int runtimeCall(@ConstantNodeParameter ForeignCallDescriptor descriptor, IsolateThread thread);
 
@@ -176,32 +199,81 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static void setHeapBase(PointerBase heapBase) {
-        if (hasHeapBase()) {
-            writeCurrentVMHeapBase(heapBase);
-            if (MemoryProtectionProvider.isAvailable()) {
-                MemoryProtectionProvider.singleton().unlockCurrentIsolate();
-            }
-        } else {
-            writeCurrentVMHeapBase(WordFactory.nullPointer());
+        writeCurrentVMHeapBase(heapBase);
+        if (MemoryProtectionProvider.isAvailable()) {
+            MemoryProtectionProvider.singleton().unlockCurrentIsolate();
         }
     }
 
     @Snippet(allowMissingProbabilities = true)
     public static int createIsolateSnippet(CEntryPointCreateIsolateParameters parameters) {
-        if (MultiThreaded.getValue()) {
-            writeCurrentVMThread(WordFactory.nullPointer());
-        }
+        writeCurrentVMThread(WordFactory.nullPointer());
         int result = runtimeCall(CREATE_ISOLATE, parameters);
         if (result != CEntryPointErrors.NO_ERROR) {
             return result;
         }
-        if (MultiThreaded.getValue()) {
-            Safepoint.transitionNativeToJava(false);
-        }
+        Safepoint.transitionNativeToJava(false);
 
         return runtimeCallInitializeIsolate(INITIALIZE_ISOLATE, parameters);
     }
 
+    private static final CGlobalData<Word> IMAGE_HEAP_PATCHING_STATE = CGlobalDataFactory.createWord();
+
+    private static final class ImageHeapPatchingState {
+        static final Word UNINITIALIZED = WordFactory.zero();
+        static final Word IN_PROGRESS = WordFactory.unsigned(1);
+        static final Word SUCCESSFUL = WordFactory.unsigned(2);
+    }
+
+    @Uninterruptible(reason = "Thread state not yet set up.")
+    private static void layeredPatchHeapRelativeRelocations() {
+        Word heapPatchStateAddr = IMAGE_HEAP_PATCHING_STATE.get();
+        boolean firstIsolate = heapPatchStateAddr.logicCompareAndSwapWord(0, ImageHeapPatchingState.UNINITIALIZED, ImageHeapPatchingState.IN_PROGRESS, NamedLocationIdentity.OFF_HEAP_LOCATION);
+
+        if (!firstIsolate) {
+            // spin-wait for first isolate
+            Word state = heapPatchStateAddr.readWordVolatile(0, LocationIdentity.ANY_LOCATION);
+            while (state.equal(ImageHeapPatchingState.IN_PROGRESS)) {
+                PauseNode.pause();
+                state = heapPatchStateAddr.readWordVolatile(0, LocationIdentity.ANY_LOCATION);
+            }
+
+            /*
+             * Patching has already been successfully completed, nothing needs to be done.
+             */
+            return;
+        }
+
+        Pointer currentSection = ImageLayerSection.getInitialLayerSection().get();
+        Word heapBegin = currentSection.readWord(ImageLayerSection.getEntryOffset(HEAP_BEGIN));
+
+        Word patches = ImageLayerSection.getHeapRelativeRelocationsStart().get();
+        int endOffset = Integer.BYTES + (patches.readInt(0) * Integer.BYTES);
+
+        int referenceSize = ConfigurationValues.getObjectLayout().getReferenceSize();
+        int offset = Integer.BYTES;
+        while (offset < endOffset) {
+            int heapOffset = patches.readInt(offset);
+            int referenceEncoding = patches.readInt(offset + Integer.BYTES);
+
+            if (referenceSize == 4) {
+                heapBegin.writeInt(heapOffset, referenceEncoding);
+            } else {
+                heapBegin.writeLong(heapOffset, referenceEncoding);
+            }
+
+            offset += 2 * Integer.BYTES;
+        }
+
+        heapPatchStateAddr.writeWordVolatile(0, ImageHeapPatchingState.SUCCESSFUL);
+    }
+
+    /**
+     * After parsing the isolate arguments in
+     * {@link IsolateArgumentParser#parse(CEntryPointCreateIsolateParameters, IsolateArguments)} the
+     * {@code providedParameters} should no longer be used. Instead {@link IsolateArguments}
+     * contains the correct values.
+     */
     @Uninterruptible(reason = "Thread state not yet set up.")
     @SubstrateForeignCallTarget(stubCallingConvention = false)
     private static int createIsolate(CEntryPointCreateIsolateParameters providedParameters) {
@@ -217,42 +289,50 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
             return CEntryPointErrors.PAGE_SIZE_CHECK_FAILED;
         }
 
+        if (ImageLayerBuildingSupport.buildingImageLayer()) {
+            layeredPatchHeapRelativeRelocations();
+        }
+
+        LocaleSupport.initialize();
+
         CEntryPointCreateIsolateParameters parameters = providedParameters;
         if (parameters.isNull() || parameters.version() < 1) {
             parameters = StackValue.get(CEntryPointCreateIsolateParameters.class);
             parameters.setReservedSpaceSize(WordFactory.zero());
             parameters.setVersion(1);
         }
-        CLongPointer parsedArgs = StackValue.get(IsolateArgumentParser.getStructSize());
-        IsolateArgumentParser.parse(parameters, parsedArgs);
-        if (parameters.reservedSpaceSize().equal(0)) {
-            parameters.setReservedSpaceSize(WordFactory.unsigned(parsedArgs.read(IsolateArgumentParser.getOptionIndex(SubstrateGCOptions.ReservedAddressSpaceSize))));
-        }
+
+        IsolateArguments arguments = StackValue.get(IsolateArguments.class);
+        UnmanagedMemoryUtil.fill((Pointer) arguments, SizeOf.unsigned(IsolateArguments.class), (byte) 0);
+        CLongPointer parsedArgs = StackValue.get(IsolateArgumentParser.getParsedArgsSize());
+        arguments.setParsedArgs(parsedArgs);
+
+        IsolateArgumentParser.singleton().parse(parameters, arguments);
+
+        Container.initialize();
 
         WordPointer isolatePtr = StackValue.get(WordPointer.class);
-        int error = Isolates.create(isolatePtr, parameters);
+        int error = Isolates.create(isolatePtr, arguments);
         if (error != CEntryPointErrors.NO_ERROR) {
+            IsolateArgumentParser.singleton().tearDown(arguments);
             return error;
         }
         Isolate isolate = isolatePtr.read();
-        if (SpawnIsolates.getValue()) {
-            setHeapBase(Isolates.getHeapBase(isolate));
-        }
+        setHeapBase(Isolates.getHeapBase(isolate));
 
-        return createIsolate0(isolate, parameters, parsedArgs);
+        return createIsolate0(isolate, arguments);
     }
 
     @Uninterruptible(reason = "Thread state not yet set up.")
     @NeverInline(value = "Ensure this code cannot rise above where heap base is set.")
-    private static int createIsolate0(Isolate isolate, CEntryPointCreateIsolateParameters parameters, CLongPointer parsedArgs) {
-        IsolateArgumentParser.persistOptions(parameters, parsedArgs);
+    private static int createIsolate0(Isolate isolate, IsolateArguments arguments) {
+        assert Heap.getHeap().verifyImageHeapMapping();
+        IsolateArgumentParser.singleton().persistOptions(arguments);
         IsolateListenerSupport.singleton().afterCreateIsolate(isolate);
 
         CodeInfoTable.prepareImageCodeInfo();
-        if (MultiThreaded.getValue()) {
-            if (!VMThreads.ensureInitialized()) {
-                return CEntryPointErrors.THREADING_INITIALIZATION_FAILED;
-            }
+        if (!VMThreads.ensureInitialized()) {
+            return CEntryPointErrors.THREADING_INITIALIZATION_FAILED;
         }
         int error = enterAttachThread0(isolate, false, true);
         if (error != CEntryPointErrors.NO_ERROR) {
@@ -310,10 +390,17 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
 
     @NeverInline("GR-24649")
     private static int initializeIsolateInterruptibly1(CEntryPointCreateIsolateParameters parameters) {
+        /* Initialize the isolate id (the id is needed for isolate teardown). */
         long initStateAddr = FIRST_ISOLATE_INIT_STATE.get().rawValue();
         boolean firstIsolate = Unsafe.getUnsafe().compareAndSetInt(null, initStateAddr, FirstIsolateInitStates.UNINITIALIZED, FirstIsolateInitStates.IN_PROGRESS);
         Isolates.assignIsolateId(firstIsolate);
 
+        /*
+         * Initialize the physical memory size. This must be done as early as possible because we
+         * must not trigger GC before PhysicalMemory is initialized.
+         */
+        PhysicalMemory.initialize();
+
         /*
          * The VM operation thread must be started early as no VM operations can be scheduled before
          * this thread is fully started. The isolate teardown may also use VM operations.
@@ -326,7 +413,8 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
          * Before this point, only limited error handling is possible because the isolate is not
          * sufficiently initialized (i.e., we can't tear it down properly).
          */
-        Isolates.setCurrentStartTime();
+        Isolates.assignStartTime();
+        LocaleSupport.checkForError();
 
         if (!firstIsolate) {
             int state = Unsafe.getUnsafe().getInt(initStateAddr);
@@ -355,6 +443,8 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
          * that allocates a significant amount of memory.
          */
 
+        IsolateArgumentParser.singleton().copyToRuntimeOptions();
+
         if (parameters.isNonNull() && parameters.version() >= 3 && parameters.getArgv().isNonNull()) {
             boolean exitWhenArgumentParsingFails = true;
             boolean ignoreUnrecognized = false;
@@ -393,9 +483,6 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
         /* Adjust stack overflow boundary of main thread. */
         StackOverflowCheck.singleton().updateStackOverflowBoundary();
 
-        /* Initialize the physical memory size. */
-        PhysicalMemory.size();
-
         assert !isolateInitialized;
         isolateInitialized = true;
 
@@ -422,17 +509,13 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
 
     @Snippet(allowMissingProbabilities = true)
     public static int attachThreadSnippet(Isolate isolate, boolean startedByIsolate, boolean ensureJavaThread) {
-        if (MultiThreaded.getValue()) {
-            writeCurrentVMThread(WordFactory.nullPointer());
-        }
+        writeCurrentVMThread(WordFactory.nullPointer());
         int error = runtimeCall(ATTACH_THREAD, isolate, startedByIsolate, ensureJavaThread);
         if (error != CEntryPointErrors.NO_ERROR) {
             return error;
         }
 
-        if (MultiThreaded.getValue()) {
-            Safepoint.transitionNativeToJava(false);
-        }
+        Safepoint.transitionNativeToJava(false);
 
         if (ensureJavaThread) {
             return runtimeCallEnsureJavaThread(ENSURE_JAVA_THREAD);
@@ -457,41 +540,35 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
         if (error != CEntryPointErrors.NO_ERROR) {
             return error;
         }
-        if (SpawnIsolates.getValue()) {
-            setHeapBase(Isolates.getHeapBase(isolate));
+        setHeapBase(Isolates.getHeapBase(isolate));
+        if (!VMThreads.isInitialized()) {
+            return CEntryPointErrors.UNINITIALIZED_ISOLATE;
         }
-        if (MultiThreaded.getValue()) {
-            if (!VMThreads.isInitialized()) {
-                return CEntryPointErrors.UNINITIALIZED_ISOLATE;
-            }
 
-            IsolateThread thread = WordFactory.nullPointer();
-            if (startedByIsolate) {
-                assert VMThreads.singleton().findIsolateThreadForCurrentOSThread(inCrashHandler).isNull();
-            } else {
-                thread = VMThreads.singleton().findIsolateThreadForCurrentOSThread(inCrashHandler);
-            }
-
-            if (thread.isNull()) { // not attached
-                if (!allowAttach) {
-                    return CEntryPointErrors.UNATTACHED_THREAD;
-                }
-                return attachUnattachedThread(isolate, startedByIsolate, inCrashHandler);
-            } else {
-                writeCurrentVMThread(thread);
-                if (ensureJavaThread && !PlatformThreads.isCurrentAssigned()) {
-                    throw VMError.shouldNotReachHere("thread was already attached but does not have a Thread object and we would assign one");
-                }
-            }
+        IsolateThread thread = WordFactory.nullPointer();
+        if (startedByIsolate) {
+            assert VMThreads.singleton().findIsolateThreadForCurrentOSThread(inCrashHandler).isNull();
         } else {
-            StackOverflowCheck.singleton().initialize();
+            thread = VMThreads.singleton().findIsolateThreadForCurrentOSThread(inCrashHandler);
+        }
+
+        if (thread.isNull()) { // not attached
+            if (!allowAttach) {
+                return CEntryPointErrors.UNATTACHED_THREAD;
+            }
+            return attachUnattachedThread(isolate, startedByIsolate, inCrashHandler);
+        } else {
+            writeCurrentVMThread(thread);
+            if (ensureJavaThread && !PlatformThreads.isCurrentAssigned()) {
+                throw VMError.shouldNotReachHere("thread was already attached but does not have a Thread object and we would assign one");
+            }
         }
         return CEntryPointErrors.NO_ERROR;
     }
 
     @Uninterruptible(reason = "Thread state not yet set up.")
     private static int attachUnattachedThread(Isolate isolate, boolean startedByIsolate, boolean inCrashHandler) {
-        int isolateThreadSize = VMThreadLocalMTSupport.singleton().vmThreadSize;
+        int isolateThreadSize = VMThreadLocalSupport.singleton().vmThreadSize;
         IsolateThread thread = VMThreads.singleton().allocateIsolateThread(isolateThreadSize);
         if (thread.isNull()) {
             return CEntryPointErrors.ALLOCATION_FAILED;
@@ -508,7 +585,7 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
         } else {
             int error = VMThreads.singleton().attachThread(thread, startedByIsolate);
             if (error != CEntryPointErrors.NO_ERROR) {
-                VMThreads.singleton().freeIsolateThread(thread);
+                VMThreads.singleton().freeCurrentIsolateThread();
                 return error;
             }
         }
@@ -529,15 +606,11 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
 
     @Uninterruptible(reason = "Thread state not yet set up.")
     public static void initializeIsolateThreadForCrashHandler(Isolate isolate, IsolateThread thread) {
-        if (SpawnIsolates.getValue()) {
-            setHeapBase(Isolates.getHeapBase(isolate));
-        }
+        setHeapBase(Isolates.getHeapBase(isolate));
 
-        if (MultiThreaded.getValue()) {
-            writeCurrentVMThread(thread);
-            VMThreads.StatusSupport.setStatusNative();
-            VMThreads.IsolateTL.set(thread, isolate);
-        }
+        writeCurrentVMThread(thread);
+        VMThreads.StatusSupport.setStatusNative();
+        VMThreads.IsolateTL.set(thread, isolate);
         SubstrateDiagnostics.setOnlyAttachedForCrashHandler(thread);
         StackOverflowCheck.singleton().disableStackOverflowChecksForFatalError();
     }
@@ -562,20 +635,15 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
     }
 
     @Snippet(allowMissingProbabilities = true)
-    public static int detachThreadSnippet() {
-        int result = CEntryPointErrors.NO_ERROR;
-        if (MultiThreaded.getValue()) {
-            IsolateThread thread = CurrentIsolate.getCurrentThread();
-            result = runtimeCall(DETACH_THREAD_MT, thread);
-        }
-        return result;
+    public static int detachCurrentThreadSnippet() {
+        return runtimeCall(DETACH_CURRENT_THREAD);
     }
 
     @SubstrateForeignCallTarget(stubCallingConvention = false)
     @Uninterruptible(reason = "Thread state going away.")
-    private static int detachThreadMT(IsolateThread currentThread) {
+    private static int detachCurrentThread() {
         try {
-            VMThreads.singleton().detachThread(currentThread);
+            VMThreads.singleton().detachCurrentThread();
         } catch (Throwable t) {
             return CEntryPointErrors.UNCAUGHT_EXCEPTION;
         }
@@ -588,46 +656,75 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
     }
 
     @SubstrateForeignCallTarget(stubCallingConvention = false)
-    @Uninterruptible(reason = "All code executed after VMThreads#tearDown must be uninterruptible")
+    @Uninterruptible(reason = "Tear-down in progress - may still execute interruptible Java code in the beginning.")
     private static int tearDownIsolate() {
         try {
+            /* Execute interruptible code. */
             if (!initiateTearDownIsolateInterruptibly()) {
                 return CEntryPointErrors.UNSPECIFIED;
             }
 
-            VMThreads.singleton().tearDown();
-            IsolateThread finalThread = CurrentIsolate.getCurrentThread();
-            int result = Isolates.tearDownCurrent();
-            // release the heap memory associated with final isolate thread
-            VMThreads.singleton().freeIsolateThread(finalThread);
-            WriteCurrentVMThreadNode.writeCurrentVMThread(WordFactory.nullPointer());
-            return result;
+            /* After threadExit(), only uninterruptible code may be executed. */
+            ThreadingSupportImpl.pauseRecurringCallback("Execution of arbitrary code is prohibited during the last teardown steps.");
+
+            /* Shut down VM thread. */
+            if (VMOperationControl.useDedicatedVMOperationThread()) {
+                VMOperationControl.shutdownAndDetachVMOperationThread();
+            }
+
+            /* Wait until all other threads exited on the OS-level. */
+            VMThreads.singleton().waitUntilDetachedThreadsExitedOnOSLevel();
+
+            IsolateThread currentThread = CurrentIsolate.getCurrentThread();
+            VMError.guarantee(VMThreads.firstThreadUnsafe().equal(currentThread));
+            VMError.guarantee(VMThreads.nextThread(VMThreads.firstThreadUnsafe()).isNull());
+
+            /* Now that we are truly single-threaded, notify listeners about isolate teardown. */
+            IsolateListenerSupport.singleton().onIsolateTeardown();
+
+            /* Free the native resources of the last thread. */
+            PlatformThreads.detach(currentThread);
+            PlatformThreads.singleton().closeOSThreadHandle(VMThreads.getOSThreadHandle(currentThread));
+
+            /* Free VM-internal native memory and tear down the Java heap. */
+            CodeInfoTable.tearDown();
+            NonmovableArrays.tearDown();
+            Heap.getHeap().tearDown();
+            IsolateArgumentParser.singleton().tearDown();
+
+            /* Tear down the heap address space, including the image heap. */
+            int code = CommittedMemoryProvider.get().tearDown();
+            if (code != CEntryPointErrors.NO_ERROR) {
+                return code;
+            }
+
+            /* Free the last thread. */
+            VMThreads.singleton().freeCurrentIsolateThread();
+            return CEntryPointErrors.NO_ERROR;
         } catch (Throwable t) {
-            return CEntryPointErrors.UNCAUGHT_EXCEPTION;
+            return reportException(t);
         }
     }
 
-    @Uninterruptible(reason = "Used as a transition between uninterruptible and interruptible code", calleeMustBe = false)
+    @Uninterruptible(reason = "Tear-down in progress - still safe to execute interruptible Java code.", calleeMustBe = false)
     private static boolean initiateTearDownIsolateInterruptibly() {
         RuntimeSupport.executeTearDownHooks();
-        return PlatformThreads.singleton().tearDown();
+        if (!PlatformThreads.tearDownOtherThreads()) {
+            return false;
+        }
+
+        VMThreads.singleton().threadExit();
+        return true;
     }
 
     @Snippet(allowMissingProbabilities = true)
     public static int enterByIsolateSnippet(Isolate isolate) {
         int result;
-        if (MultiThreaded.getValue()) {
-            writeCurrentVMThread(WordFactory.nullPointer());
-            result = runtimeCall(ENTER_BY_ISOLATE_MT, isolate);
-            if (result == CEntryPointErrors.NO_ERROR) {
-                if (VMThreads.StatusSupport.isStatusNativeOrSafepoint()) {
-                    Safepoint.transitionNativeToJava(false);
-                }
-            }
-        } else {
-            result = Isolates.checkIsolate(isolate);
-            if (result == CEntryPointErrors.NO_ERROR && SpawnIsolates.getValue()) {
-                setHeapBase(Isolates.getHeapBase(isolate));
+        writeCurrentVMThread(WordFactory.nullPointer());
+        result = runtimeCall(ENTER_BY_ISOLATE, isolate);
+        if (result == CEntryPointErrors.NO_ERROR) {
+            if (VMThreads.StatusSupport.isStatusNativeOrSafepoint()) {
+                Safepoint.transitionNativeToJava(false);
             }
         }
         return result;
@@ -635,14 +732,12 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
 
     @Uninterruptible(reason = "Thread state not set up yet")
     @SubstrateForeignCallTarget(stubCallingConvention = false)
-    private static int enterByIsolateMT(Isolate isolate) {
+    private static int enterByIsolate(Isolate isolate) {
         int error = Isolates.checkIsolate(isolate);
         if (error != CEntryPointErrors.NO_ERROR) {
             return error;
         }
-        if (SpawnIsolates.getValue()) {
-            setHeapBase(Isolates.getHeapBase(isolate));
-        }
+        setHeapBase(Isolates.getHeapBase(isolate));
         if (!VMThreads.isInitialized()) {
             return CEntryPointErrors.UNINITIALIZED_ISOLATE;
         }
@@ -656,36 +751,20 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
 
     @Snippet(allowMissingProbabilities = true)
     public static int enterSnippet(IsolateThread thread) {
-        Isolate isolate;
-        if (MultiThreaded.getValue()) {
-            if (thread.isNull()) {
-                return CEntryPointErrors.NULL_ARGUMENT;
-            }
-            writeCurrentVMThread(thread);
-            isolate = VMThreads.IsolateTL.get(thread);
-        } else { // single-threaded
-            if (SpawnIsolates.getValue()) {
-                if (thread.isNull()) {
-                    return CEntryPointErrors.NULL_ARGUMENT;
-                }
-            } else if (!thread.equal(CEntryPointSetup.SINGLE_THREAD_SENTINEL)) {
-                return CEntryPointErrors.UNATTACHED_THREAD;
-            }
-            isolate = (Isolate) ((Word) thread).subtract(CEntryPointSetup.SINGLE_ISOLATE_TO_SINGLE_THREAD_ADDEND);
+        if (thread.isNull()) {
+            return CEntryPointErrors.NULL_ARGUMENT;
         }
-        if (SpawnIsolates.getValue()) {
-            setHeapBase(Isolates.getHeapBase(isolate));
-        }
-        if (MultiThreaded.getValue()) {
-            if (runtimeAssertionsEnabled() || SubstrateOptions.CheckIsolateThreadAtEntry.getValue()) {
-                /*
-                 * Verification must happen before the thread state transition. It locks the raw
-                 * THREAD_MUTEX, so the thread must still be invisible to the safepoint manager.
-                 */
-                runtimeCall(VERIFY_ISOLATE_THREAD, thread);
-            }
-            Safepoint.transitionNativeToJava(false);
+        writeCurrentVMThread(thread);
+        Isolate isolate = VMThreads.IsolateTL.get(thread);
+        setHeapBase(Isolates.getHeapBase(isolate));
+        if (runtimeAssertionsEnabled() || SubstrateOptions.CheckIsolateThreadAtEntry.getValue()) {
+            /*
+             * Verification must happen before the thread state transition. It locks the raw
+             * THREAD_MUTEX, so the thread must still be invisible to the safepoint manager.
+             */
+            runtimeCall(VERIFY_ISOLATE_THREAD, thread);
         }
+        Safepoint.transitionNativeToJava(false);
 
         return CEntryPointErrors.NO_ERROR;
 
@@ -746,24 +825,19 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
 
     @Snippet(allowMissingProbabilities = true)
     public static int returnFromJavaToCSnippet() {
-        if (MultiThreaded.getValue()) {
-            VMThreads.StatusSupport.setStatusNative();
-        }
+        VMThreads.StatusSupport.setStatusNative();
         return CEntryPointErrors.NO_ERROR;
     }
 
     @Snippet(allowMissingProbabilities = true)
     public static boolean isAttachedSnippet(Isolate isolate) {
-        return Isolates.checkIsolate(isolate) == CEntryPointErrors.NO_ERROR &&
-                        (!MultiThreaded.getValue() || runtimeCallIsAttached(IS_ATTACHED_MT, isolate));
+        return Isolates.checkIsolate(isolate) == CEntryPointErrors.NO_ERROR && runtimeCallIsAttached(IS_ATTACHED, isolate);
     }
 
     @Uninterruptible(reason = "Thread state not yet set up.")
     @SubstrateForeignCallTarget(stubCallingConvention = false)
-    private static boolean isAttachedMT(Isolate isolate) {
-        if (SpawnIsolates.getValue()) {
-            setHeapBase(Isolates.getHeapBase(isolate));
-        }
+    private static boolean isAttached(Isolate isolate) {
+        setHeapBase(Isolates.getHeapBase(isolate));
         return VMThreads.isInitialized() && VMThreads.singleton().findIsolateThreadForCurrentOSThread(false).isNonNull();
     }
 
@@ -793,7 +867,7 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
     private final SnippetInfo enterByIsolate;
 
     private final SnippetInfo returnFromJavaToC;
-    private final SnippetInfo detachThread;
+    private final SnippetInfo detachCurrentThread;
     private final SnippetInfo reportException;
     private final SnippetInfo tearDownIsolate;
 
@@ -809,7 +883,7 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
         this.enterByIsolate = snippet(providers, CEntryPointSnippets.class, "enterByIsolateSnippet");
 
         this.returnFromJavaToC = snippet(providers, CEntryPointSnippets.class, "returnFromJavaToCSnippet");
-        this.detachThread = snippet(providers, CEntryPointSnippets.class, "detachThreadSnippet");
+        this.detachCurrentThread = snippet(providers, CEntryPointSnippets.class, "detachCurrentThreadSnippet");
         this.reportException = snippet(providers, CEntryPointSnippets.class, "reportExceptionSnippet");
         this.tearDownIsolate = snippet(providers, CEntryPointSnippets.class, "tearDownIsolateSnippet");
 
@@ -869,7 +943,7 @@ public final class CEntryPointSnippets extends SubstrateTemplates implements Sni
                     args = new Arguments(returnFromJavaToC, node.graph().getGuardsStage(), tool.getLoweringStage());
                     break;
                 case DetachThread:
-                    args = new Arguments(detachThread, node.graph().getGuardsStage(), tool.getLoweringStage());
+                    args = new Arguments(detachCurrentThread, node.graph().getGuardsStage(), tool.getLoweringStage());
                     break;
                 case TearDownIsolate:
                     args = new Arguments(tearDownIsolate, node.graph().getGuardsStage(), tool.getLoweringStage());
