@@ -30,14 +30,12 @@ import java.util.ArrayList;
 import java.util.Collections;
 import java.util.List;
 
-import org.graalvm.compiler.options.Option;
 import org.graalvm.nativeimage.ImageSingletons;
 import org.graalvm.nativeimage.Platform;
 import org.graalvm.nativeimage.Platforms;
 import org.graalvm.nativeimage.hosted.FieldValueTransformer;
 
 import com.oracle.svm.core.annotate.Alias;
-import com.oracle.svm.core.annotate.Delete;
 import com.oracle.svm.core.annotate.Inject;
 import com.oracle.svm.core.annotate.InjectAccessors;
 import com.oracle.svm.core.annotate.RecomputeFieldValue;
@@ -48,7 +46,10 @@ import com.oracle.svm.core.annotate.TargetElement;
 import com.oracle.svm.core.feature.AutomaticallyRegisteredFeature;
 import com.oracle.svm.core.feature.InternalFeature;
 import com.oracle.svm.core.option.HostedOptionKey;
+import com.oracle.svm.core.util.BasedOnJDKFile;
+import com.oracle.svm.core.util.VMError;
 
+import jdk.graal.compiler.options.Option;
 import jdk.internal.util.StaticProperty;
 
 public final class FileSystemProviderSupport {
@@ -295,6 +296,10 @@ class UnixFileSystemAccessors {
         that.injectedRootDirectory = value;
     }
 
+    @BasedOnJDKFile("https://github.com/openjdk/jdk/blob/jdk-25+20/src/java.base/linux/classes/sun/nio/fs/LinuxFileSystem.java#L44-L46")
+    @BasedOnJDKFile("https://github.com/openjdk/jdk/blob/jdk-25+20/src/java.base/linux/classes/sun/nio/fs/LinuxFileSystemProvider.java#L45-L47")
+    @BasedOnJDKFile("https://github.com/openjdk/jdk/blob/jdk-25+20/src/java.base/unix/classes/sun/nio/fs/UnixFileSystemProvider.java#L75-L77")
+    @BasedOnJDKFile("https://github.com/openjdk/jdk/blob/jdk-25+20/src/java.base/unix/classes/sun/nio/fs/UnixFileSystem.java#L78-L108")
     private static synchronized void reinitialize(Target_sun_nio_fs_UnixFileSystem that) {
         if (that.needsReinitialization != NeedsReinitializationProvider.STATUS_NEEDS_REINITIALIZATION) {
             /* Field initialized is volatile, so double-checked locking is OK. */
@@ -405,43 +410,11 @@ final class Target_java_io_UnixFileSystem {
     @Alias //
     @InjectAccessors(UserDirAccessors.class) //
     private String userDir;
-
-    @Alias //
-    @RecomputeFieldValue(kind = Kind.NewInstance, declClassName = "java.io.ExpiringCache") //
-    @TargetElement(onlyWith = JDK20OrEarlier.class)//
-    private Target_java_io_ExpiringCache cache;
-
-    /*
-     * The prefix cache on Linux/MacOS only caches elements in the Java home directory, which does
-     * not exist at image runtime. So we disable that cache completely, which is done by
-     * substituting the value of FileSystem.useCanonPrefixCache to false in the substitution below.
-     */
-    @Delete //
-    @TargetElement(onlyWith = JDK20OrEarlier.class)//
-    private String javaHome;
-    /*
-     * Ideally, we would mark this field as @Delete too. However, the javaHomePrefixCache is cleared
-     * from various methods, and we do not want to change those methods.
-     */
-    @Alias //
-    @RecomputeFieldValue(kind = Kind.NewInstance, declClassName = "java.io.ExpiringCache") //
-    @TargetElement(onlyWith = JDK20OrEarlier.class)//
-    private Target_java_io_ExpiringCache javaHomePrefixCache;
 }
 
 @TargetClass(className = "java.io.FileSystem")
 final class Target_java_io_FileSystem {
 
-    /*
-     * Linux/MacOS only: disable the usage of the javaHomePrefixCache. On Windows, the prefix cache
-     * is not specific to the Java home directory and therefore can remain enabled.
-     */
-    @Platforms({Platform.LINUX.class, Platform.DARWIN.class}) //
-    @Alias //
-    @RecomputeFieldValue(kind = Kind.FromAlias, isFinal = true) //
-    @TargetElement(onlyWith = JDK20OrEarlier.class)//
-    static boolean useCanonPrefixCache = false;
-
     @Alias
     native String normalize(String path);
 }
@@ -458,6 +431,11 @@ class UserDirAccessors {
         }
         return StaticProperty.userDir();
     }
+
+    @SuppressWarnings("unused")
+    static void setUserDir(Target_java_io_FileSystem that, String value) {
+        throw VMError.shouldNotReachHere("Field userDir is initialized at build time");
+    }
 }
 
 @TargetClass(className = "java.io.WinNTFileSystem")
@@ -467,18 +445,4 @@ final class Target_java_io_WinNTFileSystem {
     @Alias //
     @InjectAccessors(UserDirAccessors.class) //
     private String userDir;
-
-    @Alias //
-    @RecomputeFieldValue(kind = Kind.NewInstance, declClassName = "java.io.ExpiringCache") //
-    @TargetElement(onlyWith = JDK20OrEarlier.class)//
-    private Target_java_io_ExpiringCache cache;
-
-    @Alias //
-    @RecomputeFieldValue(kind = Kind.NewInstance, declClassName = "java.io.ExpiringCache") //
-    @TargetElement(onlyWith = JDK20OrEarlier.class)//
-    private Target_java_io_ExpiringCache prefixCache;
-}
-
-@TargetClass(className = "java.io.ExpiringCache", onlyWith = JDK20OrEarlier.class)
-final class Target_java_io_ExpiringCache {
 }
