@@ -24,20 +24,20 @@
  */
 package com.oracle.svm.core.jfr;
 
-import org.graalvm.compiler.api.replacements.Fold;
-import org.graalvm.nativeimage.ImageSingletons;
-
 import org.graalvm.nativeimage.c.struct.SizeOf;
-import org.graalvm.nativeimage.impl.UnmanagedMemorySupport;
 import org.graalvm.word.Pointer;
 import org.graalvm.word.UnsignedWord;
 import org.graalvm.word.WordFactory;
 
 import com.oracle.svm.core.Uninterruptible;
 import com.oracle.svm.core.config.ConfigurationValues;
+import com.oracle.svm.core.memory.NullableNativeMemory;
+import com.oracle.svm.core.nmt.NmtCategory;
 import com.oracle.svm.core.thread.VMOperation;
 import com.oracle.svm.core.util.UnsignedUtils;
 
+import jdk.graal.compiler.api.replacements.Fold;
+
 /**
  * Used to access the raw memory of a {@link JfrBuffer}.
  */
@@ -55,14 +55,14 @@ public final class JfrBufferAccess {
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static JfrBuffer allocate(JfrBufferType bufferType) {
-        long dataSize = SubstrateJVM.getThreadLocal().getThreadLocalBufferSize();
-        return allocate(WordFactory.unsigned(dataSize), bufferType);
+        UnsignedWord dataSize = SubstrateJVM.getThreadLocal().getThreadLocalBufferSize();
+        return allocate(dataSize, bufferType);
     }
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static JfrBuffer allocate(UnsignedWord dataSize, JfrBufferType bufferType) {
         UnsignedWord headerSize = JfrBufferAccess.getHeaderSize();
-        JfrBuffer result = ImageSingletons.lookup(UnmanagedMemorySupport.class).malloc(headerSize.add(dataSize));
+        JfrBuffer result = NullableNativeMemory.malloc(headerSize.add(dataSize), NmtCategory.JFR);
         if (result.isNonNull()) {
             result.setSize(dataSize);
             result.setBufferType(bufferType);
@@ -75,7 +75,7 @@ public final class JfrBufferAccess {
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static void free(JfrBuffer buffer) {
-        ImageSingletons.lookup(UnmanagedMemorySupport.class).free(buffer);
+        NullableNativeMemory.free(buffer);
     }
 
     @Uninterruptible(reason = "Prevent safepoints as those could change the flushed position.")
@@ -178,7 +178,7 @@ public final class JfrBufferAccess {
 
     /**
      * If a buffer can't be freed right away, then we retire it instead. Retired buffers are ignored
-     * by the JFR infrastructure and may be reinstated or freed at a later point in time.
+     * by the JFR infrastructure and may be reinstate or freed at a later point in time.
      */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static void setRetired(JfrBuffer buffer) {
