@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2022, 2022, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2022, 2024, Oracle and/or its affiliates. All rights reserved.
  * Copyright (c) 2022, 2022, Red Hat Inc. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
@@ -27,14 +27,16 @@
 package com.oracle.svm.core.monitor;
 
 import java.util.concurrent.TimeUnit;
+import java.util.concurrent.locks.AbstractQueuedLongSynchronizer;
 import java.util.concurrent.locks.LockSupport;
 
-import com.oracle.svm.core.NeverInline;
 import com.oracle.svm.core.Uninterruptible;
 import com.oracle.svm.core.jfr.JfrTicks;
 import com.oracle.svm.core.jfr.SubstrateJVM;
 import com.oracle.svm.core.jfr.events.JavaMonitorWaitEvent;
 import com.oracle.svm.core.thread.JavaThreads;
+import com.oracle.svm.core.util.BasedOnJDKClass;
+import com.oracle.svm.core.util.BasedOnJDKFile;
 
 import jdk.internal.misc.Unsafe;
 
@@ -43,10 +45,10 @@ import jdk.internal.misc.Unsafe;
  * {@link java.util.concurrent.locks.AbstractQueuedLongSynchronizer} as of JDK 24+11. This class
  * could be merged with {@link JavaMonitor} but we keep it separate because that way diffing against
  * the JDK sources is easier.
- * 
+ *
  * Only the relevant methods from the JDK sources have been kept. Some additional Native
  * Image-specific functionality has been added.
- * 
+ *
  * Main differences to the JDK implementation:
  * <ul>
  * <li>No need to store an owner {@linkplain Thread thread object}.</li>
@@ -57,6 +59,7 @@ import jdk.internal.misc.Unsafe;
  * {@link java.util.concurrent.ForkJoinPool#managedBlock}).</li>
  * </ul>
  */
+@BasedOnJDKClass(AbstractQueuedLongSynchronizer.class)
 abstract class JavaMonitorQueuedSynchronizer {
     // Node status bits, also used as argument and return values
     static final int WAITING = 1; // must be 1
@@ -67,6 +70,7 @@ abstract class JavaMonitorQueuedSynchronizer {
     protected static final int SPIN_SUCCESS = -1;
 
     // see AbstractQueuedLongSynchronizer.Node
+    @BasedOnJDKClass(value = AbstractQueuedLongSynchronizer.class, innerClass = "Node")
     abstract static class Node {
         volatile Node prev; // initially attached via casTail
         volatile Node next; // visibly nonnull when signallable
@@ -109,10 +113,12 @@ abstract class JavaMonitorQueuedSynchronizer {
     }
 
     // see AbstractQueuedLongSynchronizer.ExclusiveNode
+    @BasedOnJDKClass(value = AbstractQueuedLongSynchronizer.class, innerClass = "ExclusiveNode")
     static final class ExclusiveNode extends Node {
     }
 
     // see AbstractQueuedLongSynchronizer.ConditionNode
+    @BasedOnJDKClass(value = AbstractQueuedLongSynchronizer.class, innerClass = "ConditionNode")
     static final class ConditionNode extends Node {
         ConditionNode nextWaiter; // link to next waiting node
         long notifierJfrTid;
@@ -183,7 +189,7 @@ abstract class JavaMonitorQueuedSynchronizer {
             } else {
                 if (h == null) {
                     try {
-                        h = allocateExclusiveNode();
+                        h = new ExclusiveNode();
                     } catch (OutOfMemoryError oome) {
                         return null;
                     }
@@ -277,6 +283,7 @@ abstract class JavaMonitorQueuedSynchronizer {
 
     // see AbstractQueuedLongSynchronizer.acquire(Node, long, false, false, false, 0L)
     @SuppressWarnings("all")
+    @BasedOnJDKFile("https://github.com/openjdk/jdk/blob/jdk-24+24/src/hotspot/share/runtime/objectMonitor.cpp#L895-L930")
     final int acquire(Node node, long arg) {
         Thread current = Thread.currentThread();
         /* Spinning logic is SVM-specific. */
@@ -284,6 +291,15 @@ abstract class JavaMonitorQueuedSynchronizer {
         int spins = getSpinAttempts(parks);
         boolean first = false;
         Node pred = null; // predecessor of node when enqueued
+        long recheckNanos = -1;
+        if (JavaThreads.isCurrentThreadVirtualAndPinned()) {
+            /*
+             * Do not park indefinitely and instead periodically retry acquiring the monitor. This
+             * avoids liveness trouble when all carrier threads have virtual threads pinned to them
+             * and so a queued successor cannot be scheduled when it is unparked.
+             */
+            recheckNanos = 1_000_000;
+        }
 
         for (;;) {
             if (!first && (pred = (node == null) ? null : node.prev) != null && !(first = (head == pred))) {
@@ -327,7 +343,7 @@ abstract class JavaMonitorQueuedSynchronizer {
                 }
             } else if (node == null) { // allocate; retry before enqueue
                 try {
-                    node = allocateExclusiveNode();
+                    node = new ExclusiveNode();
                 } catch (OutOfMemoryError oome) {
                     return acquireOnOOME(arg);
                 }
@@ -348,7 +364,16 @@ abstract class JavaMonitorQueuedSynchronizer {
                 parks++;
                 spins = getSpinAttempts(parks);
                 try {
-                    LockSupport.park(this);
+                    if (recheckNanos == -1) {
+                        LockSupport.park(this);
+                    } else {
+                        LockSupport.parkNanos(this, recheckNanos);
+                        if (tryAcquire(arg)) {
+                            cancelAcquire(node);
+                            return 1;
+                        }
+                        recheckNanos = Math.min(recheckNanos << 3, 1_000_000_000);
+                    }
                 } catch (Error | RuntimeException ex) {
                     cancelAcquire(node); // cancel & rethrow
                     throw ex;
@@ -405,11 +430,6 @@ abstract class JavaMonitorQueuedSynchronizer {
         }
     }
 
-    @NeverInline("Can be removed once GR-51172 is resolved")
-    private static ExclusiveNode allocateExclusiveNode() {
-        return new ExclusiveNode();
-    }
-
     // see AbstractQueuedLongSynchronizer.cancelAcquire(Node, false, false)
     private int cancelAcquire(Node node) {
         if (node != null) {
@@ -448,6 +468,7 @@ abstract class JavaMonitorQueuedSynchronizer {
     }
 
     // see AbstractQueuedLongSynchronizer.ConditionObject
+    @BasedOnJDKClass(value = AbstractQueuedLongSynchronizer.class, innerClass = "ConditionObject")
     public final class JavaMonitorConditionObject {
         private transient ConditionNode firstWaiter;
         private transient ConditionNode lastWaiter;
@@ -560,7 +581,7 @@ abstract class JavaMonitorQueuedSynchronizer {
             long savedAcquisitions;
             if (tryInitializeHead() != null) {
                 try {
-                    return allocateConditionNode();
+                    return new ConditionNode();
                 } catch (OutOfMemoryError oome) {
                 }
             }
@@ -573,11 +594,6 @@ abstract class JavaMonitorQueuedSynchronizer {
             return null;
         }
 
-        @NeverInline("Can be removed once GR-51172 is resolved")
-        private static ConditionNode allocateConditionNode() {
-            return new ConditionNode();
-        }
-
         // see AbstractQueuedLongSynchronizer.ConditionObject.await()
         @SuppressWarnings("all")
         public void await(Object obj) throws InterruptedException {
