@@ -27,30 +27,42 @@ package com.oracle.svm.core.thread;
 import java.lang.Thread.UncaughtExceptionHandler;
 import java.security.AccessControlContext;
 import java.security.AccessController;
+import java.util.EnumSet;
 import java.util.concurrent.atomic.AtomicInteger;
 import java.util.concurrent.atomic.AtomicLong;
 
-import org.graalvm.compiler.api.directives.GraalDirectives;
-import org.graalvm.compiler.api.replacements.Fold;
-import org.graalvm.compiler.core.common.SuppressFBWarnings;
-import org.graalvm.compiler.replacements.ReplacementsUtil;
-import org.graalvm.compiler.serviceprovider.JavaVersionUtil;
 import org.graalvm.nativeimage.CurrentIsolate;
+import org.graalvm.nativeimage.ImageSingletons;
 import org.graalvm.nativeimage.IsolateThread;
+import org.graalvm.nativeimage.Platform;
+import org.graalvm.nativeimage.Platforms;
+import org.graalvm.nativeimage.impl.InternalPlatform;
 import org.graalvm.word.Pointer;
+import org.graalvm.word.WordFactory;
 
 import com.oracle.svm.core.AlwaysInline;
 import com.oracle.svm.core.NeverInline;
 import com.oracle.svm.core.SubstrateUtil;
 import com.oracle.svm.core.Uninterruptible;
-import com.oracle.svm.core.annotate.Alias;
-import com.oracle.svm.core.annotate.TargetClass;
+import com.oracle.svm.core.feature.AutomaticallyRegisteredImageSingleton;
+import com.oracle.svm.core.imagelayer.LastImageBuildPredicate;
+import com.oracle.svm.core.jdk.StackTraceUtils;
+import com.oracle.svm.core.layeredimagesingleton.ApplicationLayerOnlyImageSingleton;
+import com.oracle.svm.core.layeredimagesingleton.LayeredImageSingletonBuilderFlags;
+import com.oracle.svm.core.layeredimagesingleton.UnsavedSingleton;
 import com.oracle.svm.core.log.Log;
-import com.oracle.svm.core.jdk.JDK19OrLater;
 import com.oracle.svm.core.snippets.KnownIntrinsics;
 import com.oracle.svm.core.stack.StackFrameVisitor;
+import com.oracle.svm.core.threadlocal.FastThreadLocal;
+import com.oracle.svm.core.threadlocal.FastThreadLocalFactory;
+import com.oracle.svm.core.threadlocal.FastThreadLocalLong;
 import com.oracle.svm.util.ReflectionUtil;
 
+import jdk.graal.compiler.api.directives.GraalDirectives;
+import jdk.graal.compiler.core.common.SuppressFBWarnings;
+import jdk.graal.compiler.replacements.ReplacementsUtil;
+import jdk.graal.compiler.serviceprovider.JavaVersionUtil;
+
 /**
  * Implements operations on {@linkplain Target_java_lang_Thread Java threads}, which are on a higher
  * abstraction level than {@link IsolateThread}s. This class distinguishes these types of threads:
@@ -68,16 +80,47 @@ import com.oracle.svm.util.ReflectionUtil;
  * Methods with <em>platform</em> or <em>carrier</em> in their names must be called <em>only</em>
  * for that type of thread. Methods without that designation distinguish between the thread types
  * and choose the appropriate action.
- *
- * @see VirtualThreads
- * @see <a href="https://openjdk.java.net/projects/loom/">Wiki and source code of Project Loom on
- *      which concepts of virtual threads, carrier threads, etc. are modeled</a>
  */
 public final class JavaThreads {
-    /** For Thread.nextThreadID(). */
-    static final AtomicLong threadSeqNumber = new AtomicLong();
-    /** For Thread.nextThreadNum(). */
-    static final AtomicInteger threadInitNumber = new AtomicInteger();
+    /**
+     * The {@linkplain JavaThreads#getThreadId thread id} of the {@link Thread#currentThread()},
+     * which can be a {@linkplain Target_java_lang_Thread#vthread virtual thread} or a
+     * {@linkplain PlatformThreads#currentThread platform thread}.
+     *
+     * As the value of the thread local can change over the thread lifetime (see carrier threads),
+     * it should only be accessed by the owning thread (via {@link FastThreadLocalLong#get()} and
+     * {@link FastThreadLocalLong#set(long)}).
+     */
+    static final FastThreadLocalLong currentVThreadId = FastThreadLocalFactory.createLong("JavaThreads.currentVThreadId").setMaxOffset(FastThreadLocal.BYTE_OFFSET);
+
+    @AutomaticallyRegisteredImageSingleton(onlyWith = LastImageBuildPredicate.class)
+    public static class JavaThreadNumberSingleton implements ApplicationLayerOnlyImageSingleton, UnsavedSingleton {
+
+        public static JavaThreadNumberSingleton singleton() {
+            return ImageSingletons.lookup(JavaThreadNumberSingleton.class);
+        }
+
+        /** For Thread.nextThreadID(). */
+        final AtomicLong threadSeqNumber;
+        /** For Thread.nextThreadNum(). */
+        final AtomicInteger threadInitNumber;
+
+        public JavaThreadNumberSingleton() {
+            this.threadSeqNumber = new AtomicLong();
+            this.threadInitNumber = new AtomicInteger();
+        }
+
+        @Platforms(Platform.HOSTED_ONLY.class)
+        public void setThreadNumberInfo(long seqNumber, int initNumber) {
+            this.threadSeqNumber.set(seqNumber);
+            this.threadInitNumber.set(initNumber);
+        }
+
+        @Override
+        public EnumSet<LayeredImageSingletonBuilderFlags> getImageBuilderFlags() {
+            return LayeredImageSingletonBuilderFlags.ALL_ACCESS;
+        }
+    }
 
     private JavaThreads() {
     }
@@ -93,6 +136,16 @@ public final class JavaThreads {
         return Target_java_lang_Thread.class.cast(thread);
     }
 
+    @Platforms(InternalPlatform.NATIVE_ONLY.class)
+    static long nextThreadID() {
+        return JavaThreadNumberSingleton.singleton().threadSeqNumber.incrementAndGet();
+    }
+
+    @Platforms(InternalPlatform.NATIVE_ONLY.class)
+    static int nextThreadNum() {
+        return JavaThreadNumberSingleton.singleton().threadInitNumber.incrementAndGet();
+    }
+
     /**
      * Returns the unique identifier of this thread. This method is necessary because
      * {@code Thread#getId()} is a non-final method that can be overridden.
@@ -132,62 +185,43 @@ public final class JavaThreads {
         return toTarget(thread).interrupted;
     }
 
-    static boolean getAndClearInterrupt(Thread thread) {
-        if (supportsVirtual() && isVirtual(thread)) {
-            return VirtualThreads.singleton().getAndClearInterrupt(thread);
-        }
-        /*
-         * As we don't use a lock, it is possible to observe any kinds of races with other threads
-         * that try to set the interrupted status to true. However, those races don't cause any
-         * correctness issues as we only reset it to false if we observed that it was true earlier.
-         * There also can't be any problematic races with other calls to check the interrupt status
-         * because it is cleared only by the current thread.
-         */
-        if (!isInterrupted(thread)) {
-            return false;
-        }
-        toTarget(thread).interrupted = false;
-        return true;
-    }
-
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static long getParentThreadId(Thread thread) {
         return toTarget(thread).parentThreadId;
     }
 
-    @Fold
-    static boolean supportsVirtual() {
-        return VirtualThreads.isSupported();
-    }
-
+    /**
+     * Indicates whether a thread is <em>truly</em> virtual, whereas {@link Thread#isVirtual()} also
+     * returns {@code true} for platform threads of type {@code BoundVirtualThread}.
+     */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static boolean isVirtual(Thread thread) {
-        return supportsVirtual() && VirtualThreads.singleton().isVirtual(thread);
+        return Target_java_lang_VirtualThread.class.isInstance(thread);
     }
 
-    /** @see PlatformThreads#setCurrentThread */
+    /** @see #isVirtual */
     public static boolean isCurrentThreadVirtual() {
-        if (!supportsVirtual()) {
-            return false;
-        }
         Thread thread = PlatformThreads.currentThread.get();
         return thread != null && toTarget(thread).vthread != null;
     }
 
+    /**
+     * Indicates whether the current thread is <em>truly</em> virtual (see {@link #isVirtual}) and
+     * currently pinned to its carrier thread.
+     */
+    public static boolean isCurrentThreadVirtualAndPinned() {
+        Target_java_lang_Thread carrier = JavaThreads.toTarget(Target_java_lang_Thread.currentCarrierThread());
+        return carrier != null && carrier.vthread != null && Target_jdk_internal_vm_Continuation.isPinned(carrier.cont.getScope());
+    }
+
     @SuppressFBWarnings(value = "BC", justification = "Cast for @TargetClass")
     static Target_java_lang_ThreadGroup toTarget(ThreadGroup threadGroup) {
         return Target_java_lang_ThreadGroup.class.cast(threadGroup);
     }
 
-    static void join(Thread thread, long millis) throws InterruptedException {
-        if (millis < 0) {
-            throw new IllegalArgumentException("Timeout value is negative");
-        }
-        if (supportsVirtual() && isVirtual(thread)) {
-            VirtualThreads.singleton().join(thread, millis);
-        } else {
-            PlatformThreads.join(thread, millis);
-        }
+    @SuppressFBWarnings(value = "BC", justification = "Cast for @TargetClass")
+    private static Target_java_lang_VirtualThread toVirtualTarget(Thread thread) {
+        return Target_java_lang_VirtualThread.class.cast(thread);
     }
 
     @NeverInline("Starting a stack walk in the caller frame")
@@ -199,10 +233,55 @@ public final class JavaThreads {
          */
         Pointer callerSP = KnownIntrinsics.readCallerStackPointer();
 
-        if (supportsVirtual()) { // NOTE: also for platform threads!
-            return VirtualThreads.singleton().getVirtualOrPlatformThreadStackTrace(filterExceptions, thread, callerSP);
+        if (isVirtual(thread)) {
+            if (thread == Thread.currentThread()) {
+                return getMountedVirtualThreadStackTrace(filterExceptions, thread, callerSP);
+            }
+            assert !filterExceptions : "exception stack traces can be taken only for the current thread";
+            return asyncGetVirtualThreadStackTrace(toVirtualTarget(thread));
+        } else {
+            return PlatformThreads.getStackTrace(filterExceptions, thread, callerSP);
         }
-        return PlatformThreads.getStackTrace(filterExceptions, thread, callerSP);
+    }
+
+    private static StackTraceElement[] getMountedVirtualThreadStackTrace(boolean filterExceptions, Thread thread, Pointer callerSP) {
+        Thread carrier = toVirtualTarget(thread).carrierThread;
+        if (carrier == null) {
+            return null;
+        }
+        Pointer endSP = PlatformThreads.getCarrierSPOrElse(carrier, WordFactory.nullPointer());
+        if (endSP.isNull()) {
+            return null;
+        }
+        if (carrier == PlatformThreads.currentThread.get()) {
+            return StackTraceUtils.getCurrentThreadStackTrace(filterExceptions, callerSP, endSP);
+        }
+        assert VMOperation.isInProgressAtSafepoint();
+        return StackTraceUtils.getStackTraceAtSafepoint(PlatformThreads.getIsolateThread(carrier), endSP);
+    }
+
+    public static StackTraceElement[] getStackTraceAtSafepoint(Thread thread, Pointer callerSP) {
+        if (isVirtual(thread)) {
+            return getMountedVirtualThreadStackTrace(false, thread, callerSP);
+        } else {
+            return PlatformThreads.getStackTraceAtSafepoint(thread, callerSP);
+        }
+    }
+
+    /** Adapted from {@code VirtualThread.asyncGetStackTrace()}. */
+    private static StackTraceElement[] asyncGetVirtualThreadStackTrace(Target_java_lang_VirtualThread thread) {
+        StackTraceElement[] stackTrace;
+        do {
+            if (thread.carrierThread != null) {
+                stackTrace = StackTraceUtils.asyncGetStackTrace(SubstrateUtil.cast(thread, Thread.class));
+            } else {
+                stackTrace = thread.tryGetStackTrace();
+            }
+            if (stackTrace == null) {
+                Thread.yield();
+            }
+        } while (stackTrace == null);
+        return stackTrace;
     }
 
     @NeverInline("Starting a stack walk in the caller frame")
@@ -214,13 +293,23 @@ public final class JavaThreads {
          */
         Pointer callerSP = KnownIntrinsics.readCallerStackPointer();
 
-        if (supportsVirtual()) { // NOTE: also for platform threads!
-            VirtualThreads.singleton().visitCurrentVirtualOrPlatformThreadStackFrames(callerSP, visitor);
+        if (isVirtual(Thread.currentThread())) {
+            visitCurrentVirtualThreadStackFrames(callerSP, visitor);
         } else {
             PlatformThreads.visitCurrentStackFrames(callerSP, visitor);
         }
     }
 
+    private static void visitCurrentVirtualThreadStackFrames(Pointer callerSP, StackFrameVisitor visitor) {
+        Thread carrier = toVirtualTarget(Thread.currentThread()).carrierThread;
+        if (carrier != null) {
+            Pointer endSP = PlatformThreads.getCarrierSPOrElse(carrier, WordFactory.nullPointer());
+            if (endSP.isNonNull()) {
+                StackTraceUtils.visitCurrentThreadStackFrames(callerSP, endSP, visitor);
+            }
+        }
+    }
+
     public static void dispatchUncaughtException(Thread thread, Throwable throwable) {
         try {
             /* Get the uncaught exception handler for the Thread, or the default one. */
@@ -270,7 +359,6 @@ public final class JavaThreads {
                     String name,
                     long stackSize,
                     AccessControlContext acc,
-                    boolean allowThreadLocals,
                     boolean inheritThreadLocals) {
         if (name == null) {
             throw new NullPointerException("The name cannot be null");
@@ -292,65 +380,30 @@ public final class JavaThreads {
 
         initThreadFields(tjlt, group, target, stackSize, priority, daemon);
 
-        if (!VirtualThreads.isSupported() || !VirtualThreads.singleton().isVirtual(fromTarget(tjlt))) {
-            PlatformThreads.setThreadStatus(fromTarget(tjlt), ThreadStatus.NEW);
-            if (JavaVersionUtil.JAVA_SPEC < 19) {
-                JavaThreads.toTarget(group).addUnstarted();
-            }
+        PlatformThreads.setThreadStatus(fromTarget(tjlt), ThreadStatus.NEW);
+
+        if (JavaVersionUtil.JAVA_SPEC == 21) {
+            tjlt.inheritedAccessControlContext = acc != null ? acc : AccessController.getContext();
         }
 
-        tjlt.inheritedAccessControlContext = acc != null ? acc : AccessController.getContext();
-
-        initNewThreadLocalsAndLoader(tjlt, allowThreadLocals, inheritThreadLocals, parent);
+        initNewThreadLocalsAndLoader(tjlt, inheritThreadLocals, parent);
 
         /* Set thread ID */
-        tjlt.tid = Target_java_lang_Thread.nextThreadID();
+        tjlt.tid = nextThreadID();
     }
 
     static void initThreadFields(Target_java_lang_Thread tjlt, ThreadGroup group, Runnable target, long stackSize, int priority, boolean daemon) {
-        if (JavaVersionUtil.JAVA_SPEC >= 19) {
-            assert tjlt.holder == null;
-            tjlt.holder = new Target_java_lang_Thread_FieldHolder(group, target, stackSize, priority, daemon);
-        } else {
-            tjlt.group = group;
-            tjlt.priority = priority;
-            tjlt.daemon = daemon;
-            tjlt.target = target;
-            tjlt.setPriority(priority);
-
-            /* Stash the specified stack size in case the VM cares */
-            tjlt.stackSize = stackSize;
-        }
+        assert tjlt.holder == null;
+        tjlt.holder = new Target_java_lang_Thread_FieldHolder(group, target, stackSize, priority, daemon);
     }
 
-    static void initNewThreadLocalsAndLoader(Target_java_lang_Thread tjlt, boolean allowThreadLocals, boolean inheritThreadLocals, Thread parent) {
-        if (JavaVersionUtil.JAVA_SPEC >= 19 && JavaVersionUtil.JAVA_SPEC <= 20 && !allowThreadLocals) {
-            tjlt.threadLocals = Target_java_lang_ThreadLocal_ThreadLocalMap.NOT_SUPPORTED;
-            tjlt.inheritableThreadLocals = Target_java_lang_ThreadLocal_ThreadLocalMap.NOT_SUPPORTED;
-            tjlt.contextClassLoader = Target_java_lang_Thread_Constants.NOT_SUPPORTED_CLASSLOADER;
-        } else if (inheritThreadLocals) {
+    static void initNewThreadLocalsAndLoader(Target_java_lang_Thread tjlt, boolean inheritThreadLocals, Thread parent) {
+        if (inheritThreadLocals) {
             Target_java_lang_ThreadLocal_ThreadLocalMap parentMap = toTarget(parent).inheritableThreadLocals;
-            if (parentMap != null) {
-                boolean inherit = false;
-                if (JavaVersionUtil.JAVA_SPEC < 19) {
-                    inherit = true;
-                } else if (parentMap.size() > 0) {
-                    if (JavaVersionUtil.JAVA_SPEC > 20) {
-                        inherit = true;
-                    } else {
-                        inherit = parentMap != Target_java_lang_ThreadLocal_ThreadLocalMap.NOT_SUPPORTED;
-                    }
-                }
-                if (inherit) {
-                    tjlt.inheritableThreadLocals = Target_java_lang_ThreadLocal.createInheritedMap(parentMap);
-                }
-            }
-            ClassLoader parentLoader = parent.getContextClassLoader();
-            if (JavaVersionUtil.JAVA_SPEC < 19 || JavaVersionUtil.JAVA_SPEC > 20 || parentLoader != Target_java_lang_Thread_Constants.NOT_SUPPORTED_CLASSLOADER) {
-                tjlt.contextClassLoader = parentLoader;
-            } else {
-                tjlt.contextClassLoader = ClassLoader.getSystemClassLoader();
+            if (parentMap != null && parentMap.size() > 0) {
+                tjlt.inheritableThreadLocals = Target_java_lang_ThreadLocal.createInheritedMap(parentMap);
             }
+            tjlt.contextClassLoader = parent.getContextClassLoader();
         } else {
             tjlt.contextClassLoader = ClassLoader.getSystemClassLoader();
         }
@@ -358,25 +411,20 @@ public final class JavaThreads {
 
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static void setCurrentThreadLockHelper(Object helper) {
-        if (supportsVirtual()) {
-            toTarget(Thread.currentThread()).lockHelper = helper;
-        } else {
-            PlatformThreads.lockHelper.set(helper);
-        }
+        toTarget(Thread.currentThread()).lockHelper = helper;
     }
 
     @AlwaysInline("Locking fast path.")
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static Object getCurrentThreadLockHelper() {
-        if (supportsVirtual()) {
-            return toTarget(Thread.currentThread()).lockHelper;
-        }
-        return PlatformThreads.lockHelper.get();
+        return toTarget(Thread.currentThread()).lockHelper;
     }
 
-    public static void blockedOn(Target_sun_nio_ch_Interruptible b) {
-        if (supportsVirtual() && isCurrentThreadVirtual()) {
-            VirtualThreads.singleton().blockedOn(b);
+    static void blockedOn(Target_sun_nio_ch_Interruptible b) {
+        assert JavaVersionUtil.JAVA_SPEC <= 21 : "blockedOn in newer JDKs uses safe disableSuspendAndPreempt";
+
+        if (isCurrentThreadVirtual()) {
+            VirtualThreadHelper.blockedOn(b);
         } else {
             PlatformThreads.blockedOn(b);
         }
@@ -388,7 +436,7 @@ public final class JavaThreads {
      */
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static long getCurrentThreadId() {
-        long id = PlatformThreads.currentVThreadId.get();
+        long id = currentVThreadId.get();
         if (GraalDirectives.inIntrinsic()) {
             ReplacementsUtil.dynamicAssert(id != 0 && id == getThreadId(Thread.currentThread()), "ids must match");
         } else {
@@ -406,11 +454,19 @@ public final class JavaThreads {
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static long getCurrentThreadIdOrZero() {
         if (CurrentIsolate.getCurrentThread().isNonNull()) {
-            return PlatformThreads.currentVThreadId.get();
+            return currentVThreadId.get();
         }
         return 0L;
     }
 
+    @Uninterruptible(reason = "Ensure consistency of vthread and cached vthread id.")
+    static void setCurrentThread(Thread carrier, Thread thread) {
+        assert carrier == PlatformThreads.currentThread.get();
+        assert thread == carrier || isVirtual(thread);
+        toTarget(carrier).vthread = (thread != carrier) ? thread : null;
+        currentVThreadId.set(getThreadId(thread));
+    }
+
     @Uninterruptible(reason = "Called from uninterruptible code.", mayBeInlined = true)
     public static Thread getCurrentThreadOrNull() {
         Thread thread = PlatformThreads.currentThread.get();
@@ -422,18 +478,3 @@ public final class JavaThreads {
         return (tjlt.vthread != null) ? tjlt.vthread : thread;
     }
 }
-
-/* GR-43733: this class can be removed when we drop the JDK 17 support. */
-@TargetClass(className = "jdk.internal.event.ThreadSleepEvent", onlyWith = JDK19OrLater.class)
-final class Target_jdk_internal_event_ThreadSleepEvent {
-    @Alias public long time;
-
-    @Alias
-    public static native boolean isTurnedOn();
-
-    @Alias
-    public native void begin();
-
-    @Alias
-    public native void commit();
-}
